from typing import Any
from torch.utils.data import Dataset
import os
import torchvision.datasets.utils
import pandas as pd
from .utils import map_column
import scipy.io as scio

class RatingPredictDataset(Dataset):
    urls = {}
    filenames = {}
    def __init__(self, root:str, download:bool=True, transform:Any=None, split:str=''):
        if split not in self.urls.keys(): raise ValueError("{} is not in list {}".format(split, list(self.urls.keys())))
        self.root = root
        self.enum_data = []
        self.unenum_data = []
        self.transform = transform
        self.split = split
        self.num_users = 0
        self.num_items = 0
        if download:
            self.download()
        self.load_data()

    def get_unenum_data(self, key=None):
        if key is not None:
            if key not in self.unenum_data: return {}
            return {key: getattr(self, key)}
        return {k:getattr(self, k) for k in self.unenum_data}

    def download(self):
        filename = os.path.join(self.root, self.filenames[self.split])
        urls = self.urls[self.split]
        if not os.path.exists(filename):
            for url in urls:
                if url.endswith('.zip' or '.tar.gz'):
                    torchvision.datasets.utils.download_and_extract_archive(url, download_root=self.root, remove_finished=True)
                else:
                    torchvision.datasets.utils.download_url(url, root=self.root)

    def __len__(self):
        if len(self.enum_data)==0:return 0
        return len(getattr(self, self.enum_data[0]))

    def __getitem__(self, item):
        data = {attr: getattr(self, attr)[item] for attr in self.enum_data}
        data.update({attr: getattr(self, attr) for attr in self.unenum_data})
        data.update({'__enum__':self.enum_data})
        return data

    def load_data(self):
        """load original data and set self.enum_data and self.unenum_data"""
        return NotImplementedError

class MovieLens(RatingPredictDataset):
    urls = {
        '100k': ["https://files.grouplens.org/datasets/movielens/ml-100k.zip"],
        '1m': ["https://files.grouplens.org/datasets/movielens/ml-1m.zip"],
        '10m': ["https://files.grouplens.org/datasets/movielens/ml-10m.zip"],
        '20m': ["https://files.grouplens.org/datasets/movielens/ml-20m.zip"],
        '25m': ["https://files.grouplens.org/datasets/movielens/ml-25m.zip"],
        'latest-small': ["https://files.grouplens.org/datasets/movielens/ml-latest-small.zip"],
        'latest-full': ["https://files.grouplens.org/datasets/movielens/ml-latest.zip"],
    }
    filenames = {
        '100k': "ml-100k",
        '1m': "ml-1m",
        '10m': "ml-10m",
        '20m': "ml-20m",
        '25m': "ml-25m",
        'latest-small': "ml-latest-small",
        'latest-full': "ml-latest",
    }
    def __init__(self, root:str, split:str='100k', download:bool=True, train:bool=True, min_val:int=10, max_val:int=10e8):
        self.train = train
        self.min_val = min_val
        self.max_val = max_val
        super(MovieLens, self).__init__(root=root, split=split, download=download)


    def load_data(self):
        if self.split == '100k':
            rnames = ['user_id', 'movie_id', 'rating', 'timestamp']
            file = 'u.data'
            data = pd.read_csv(os.path.join(self.root, self.filenames[self.split], file), sep='\t', names=rnames).drop(columns=['timestamp']).astype(int)
        else:
            rnames = ['user_id', 'movie_id', 'rating', 'timestamp']
            data = pd.read_table(os.path.join(self.root, self.filenames[self.split],'ratings.dat'), sep='::', header=None, names=rnames, engine='python').drop(columns=['timestamp']).astype(int)
        # filter the data according to the numbers of samples of users
        data = data.groupby('user_id').filter(lambda x: len(x)>self.min_val and len(x)<self.max_val)
        data['user_id'] = map_column(data, 'user_id')
        data['movie_id'] = map_column(data, 'movie_id')
        self.num_users = len(data['user_id'].unique())
        self.num_items = len(data['movie_id'].unique())
        # split the training part and the testing part
        if self.train:
            data = data.groupby('user_id').apply(lambda x: x.head(int(len(x)*0.8)))
        else:
            data = data.groupby('user_id').apply(lambda x: x.tail(len(x)-int(len(x)*0.8)))
        data = data.rename(columns = {'movie_id':'item_id'})
        enum_data = ['user_id', 'item_id', 'rating']
        self.enum_data = enum_data
        for k in self.enum_data:
            setattr(self, k, data[k].to_numpy())
        return

class Ciao(RatingPredictDataset):
    urls = {
        'train':["https://www.cse.msu.edu/~tangjili/datasetcode/ciao.zip"],
        'val':["https://www.cse.msu.edu/~tangjili/datasetcode/ciao.zip"]
    }
    filenames = {
        'train':"ciao",
        'val':"ciao"
    }
    def __init__(self, root:str, split:str='train', download:bool=True, min_val:int=0, max_val:int=10e8):
        self.min_val = min_val
        self.max_val = max_val
        super(Ciao, self).__init__(root=root, split=split, download=download)

    def load_data(self):
        # read data into memory
        records = scio.loadmat(os.path.join(self.root, 'ciao', 'rating.mat'))['rating']
        data = pd.DataFrame(records, columns=['user_id', 'item_id', 'item_category', 'rating', 'rating_conf'])
        data = data.groupby('user_id').filter(lambda x: len(x) > self.min_val and len(x) < self.max_val)
        original_users = data['user_id'].unique()
        links = pd.DataFrame(scio.loadmat(os.path.join(self.root, 'ciao', 'trustnetwork.mat'))['trustnetwork'], columns=['trustor', 'trustee'])
        links = links[links['trustor'].isin(original_users) & links['trustee'].isin(original_users)]
        map_dict ={k:v for k,v in zip(original_users, list(range(len(original_users))))}
        links['trustor'] = map_column(links, 'trustor', [map_dict[k] for k in links['trustor'].unique()])
        links['trustee'] = map_column(links, 'trustee', [map_dict[k] for k in links['trustee'].unique()])
        data['user_id'] = map_column(data, 'user_id')
        data['item_id'] = map_column(data, 'item_id')
        self.num_users = len(data['user_id'].unique())
        self.num_items = len(data['item_id'].unique())
        self.enum_data = ['user_id', 'item_id', 'rating', 'item_category',  'rating_conf']
        for k in self.enum_data: setattr(self, k, data[k].to_numpy())
        self.unenum_data = ['social_link']
        setattr(self, 'social_link', links.to_numpy())
        return

class Epinions(RatingPredictDataset):
    urls = {
        'train': ["https://www.cse.msu.edu/~tangjili/datasetcode/epinions.zip"],
        'val': ["https://www.cse.msu.edu/~tangjili/datasetcode/epinions.zip"]
    }
    filenames = {
        'train': "epinions",
        'val': "epinions"
    }

    def __init__(self, root: str, split: str = 'train', download: bool = True, min_val: int = 0, max_val: int = 10e8):
        self.min_val = min_val
        self.max_val = max_val
        super(Epinions, self).__init__(root=root, split=split, download=download)

    def load_data(self):
        # read data into memory
        records = scio.loadmat(os.path.join(self.root, 'epinions', 'rating.mat'))['rating']
        data = pd.DataFrame(records, columns=['user_id', 'item_id', 'item_category', 'rating'])
        data = data.groupby('user_id').filter(lambda x: len(x) > self.min_val and len(x) < self.max_val)
        original_users = data['user_id'].unique()
        links = pd.DataFrame(scio.loadmat(os.path.join(self.root, 'epinions', 'trustnetwork.mat'))['trustnetwork'],
                             columns=['trustor', 'trustee'])
        links = links[links['trustor'].isin(original_users) & links['trustee'].isin(original_users)]
        map_dict = {k: v for k, v in zip(original_users, list(range(len(original_users))))}
        links['trustor'] = map_column(links, 'trustor', [map_dict[k] for k in links['trustor'].unique()])
        links['trustee'] = map_column(links, 'trustee', [map_dict[k] for k in links['trustee'].unique()])
        data['user_id'] = map_column(data, 'user_id')
        data['item_id'] = map_column(data, 'item_id')
        self.num_users = len(data['user_id'].unique())
        self.num_items = len(data['item_id'].unique())
        self.enum_data = ['user_id', 'item_id', 'rating', 'item_category']
        for k in self.enum_data: setattr(self, k, data[k].to_numpy())
        self.unenum_data = ['social_link']
        setattr(self, 'social_link', links.to_numpy())
        return


from flgo.benchmark.vertical_mnist_classification.model import mlp
import flgo.benchmark.toolkits.partition

default_partitioner = flgo.benchmark.toolkits.partition.VerticalSplittedPartitioner
default_partition_para = {'num_parties':3}
default_model = mlp


import torch
import torchvision.models
import torchvision.transforms as T
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from flgo.benchmark.toolkits.cv.detection import *
import torchvision.datasets
import os

CLASSES = (
        '__background__',
        "aeroplane",
        "bicycle",
        "bird",
        "boat",
        "bottle",
        "bus",
        "car",
        "cat",
        "chair",
        "cow",
        "diningtable",
        "dog",
        "horse",
        "motorbike",
        "person",
        "pottedplant",
        "sheep",
        "sofa",
        "train",
        "tvmonitor",
    )
CLASSES_MAP = {name:idx for idx, name in enumerate(CLASSES)}
def voc_target_transform(y):
    objects = y['annotation']['object']
    boxes = [torch.FloatTensor([int(v) for v in obj['bndbox'].values()]) for obj in objects]
    labels = [torch.LongTensor(torch.LongTensor([CLASSES_MAP[obj['name'].lower()]])) for obj in objects]
    return {'boxes': torch.stack(boxes), 'labels':torch.cat(labels)}

transform = T.Compose([T.ToTensor(), T.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])])
path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'RAW_DATA', 'PASCAL_VOC')
train_data = torchvision.datasets.VOCDetection(root=path, download=True, image_set='trainval', year='2007', transform=transform, target_transform=voc_target_transform)
test_data = torchvision.datasets.VOCDetection(root=path, download=True, image_set='test', year='2007', transform=transform, target_transform=voc_target_transform)
train_data.num_classes = len(CLASSES)
test_data.num_classes = len(CLASSES)

def compute_loss(model_output, *args, **kwargs):
    return sum(list(model_output.values()))

def get_model():
    model = torchvision.models.detection.fasterrcnn_resnet50_fpn(weights=torchvision.models.detection.FasterRCNN_ResNet50_FPN_Weights.COCO_V1)
    in_features = model.roi_heads.box_predictor.cls_score.in_features
    model.roi_heads.box_predictor = FastRCNNPredictor(in_features, len(CLASSES))
    model.compute_loss = compute_loss
    return model




import torch.nn
import flgo.benchmark
import os
import torch_geometric
from torch_geometric.nn import GCNConv
import torch.nn.functional as F

class GCN(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = GCNConv(3703, 16)
        self.conv2 = GCNConv(16,6)

    def forward(self, data):
        x, edge_index = data.x, data.edge_index
        x = self.conv1(x, edge_index)
        x = F.relu(x)
        x = F.dropout(x, training=self.training)
        x = self.conv2(x, edge_index)
        return F.log_softmax(x, dim=1)

path = os.path.join(flgo.benchmark.data_root, 'CITESEER')
dataset = torch_geometric.datasets.Planetoid(path, name='Citeseer')
train_data = dataset[0]

def get_model():
    return GCN()


import os
import math
import copy
import collections
from typing import Any, Callable
from tqdm import tqdm
import torch
import torch.multiprocessing as mp
import multiprocessing.dummy as dmp
import numpy as np
import flgo.benchmark.base
from flgo.utils import fmodule
import flgo.simulator.base as ss
# import threading
# from concurrent.futures import ThreadPoolExecutor,wait,ALL_COMPLETED,FIRST_COMPLETED, as_completed

# The BasicParty
class BasicParty:
    def __init__(self, *args, **kwargs):
        self.actions = {}  # the message-action map that is used to customize the communication process
        self.id = None  # the id for communicating
        self._object_map = {} # mapping objects according to their ids
        self._data_names = []

    def register_action_to_mtype(self, action_name: str, mtype):
        r"""
        Register an existing method as the action corresponding to the message type.

        Args:
            action_name: the name of the instance method
            mtype: the message type
        """
        if action_name not in self.__dict__.keys():
            raise NotImplementedError("There is no method named `{}` in the class instance.".format(action_name))
        self.actions[mtype] = self.__dict__[action_name]

    def message_handler(self, package):
        r"""
        Handling the received message by excuting the corresponding action.

        Args:
            package (dict): the package received from other parties (i.e. the content of the message)

        Returns:
            action_reult
        """
        try:
            mtype = package['__mtype__']
        except:
            raise KeyError("__mtype__ must be a key of the package")
        if mtype not in self.actions.keys():
            raise NotImplementedError("There is no action corresponding to message type {}.".format(mtype))
        return self.actions[mtype](package)

    def set_data(self, data, flag:str='train') -> None:
        r"""
        Set self's attibute 'xxx_data' to be data where xxx is the flag. For example,
        after calling self.set_data([1,2,3], 'test'), self.test_data will be [1,2,3].
        Particularly, If the flag is 'train', the batchsize and the num_steps will be
        reset.

        Args:
            data: anything
            flag (str): the name of the data
        """
        setattr(self, flag + '_data', data)
        if flag not in self._data_names:
            self._data_names.append(flag)
        if data is None: return
        if flag == 'train' and 'real' not in self.option['scene']:
            self.datavol = len(data)
            if hasattr(self, 'batch_size'):
                # reset batch_size
                if self.batch_size < 0:
                    self.batch_size = len(self.get_data(flag))
                elif self.batch_size >= 1:
                    self.batch_size = int(self.batch_size)
                else:
                    self.batch_size = int(self.datavol * self.batch_size)
            # reset num_steps
            if hasattr(self, 'num_steps') and hasattr(self, 'num_epochs'):
                if self.num_steps > 0:
                    self.num_epochs = 1.0 * self.num_steps / (math.ceil(self.datavol / self.batch_size))
                else:
                    self.num_steps = self.num_epochs * math.ceil(self.datavol / self.batch_size)

    def get_data(self, flag:str='val')->Any:
        r"""
        Get self's attibute '{flag}_data' if this attribute exists.

        Args:
            flag (str): the name of the data

        Returns:
            flag_data (Any): self.{flag}_data
        """
        dname = (flag+'_data')
        if flag not in self._data_names: return None
        else:
            return getattr(self, dname)

    def get_data_names(self)->list:
        """
        Get the names of data hold by self.

        Returns:
            data_names (list): the names of data hold by self
        """
        return self._data_names

    def get_classname(self)->str:
        """
        Get the class name of self.

        Returns:
            class_name (str): the class name
        """
        return self.__class__.__name__

    def set_model(self, model, model_name: str = 'model'):
        r"""
        Set self's attibute 'model_name' to be model. For example,
        after calling self.set_model(my_model, 'model'), self.model will be my_model.
        """
        # set self.__dict__[model_name] = model
        setattr(self, model_name, model)

    def set_id(self, id=None):
        r"""
        Set self's attibute 'id' to be id where self.id = id
        """
        if id is not None:
            self.id = id

    def set_message(self, mtype:Any, package:dict={})->dict:
        """Set the message type of a package.

        Args:
            mtype (Any): the message type
            package (dict): a dict

        Returns:
            package_with_mtype (dict): a dict with the message type
        """
        if type(package) is not dict:
            raise TypeError('The type of the package should be dict')
        package.update({'__mtype__': mtype})
        return package

    def register_objects(self, parties:list, parties_name='parties'):
        r"""
        Set self's attribute party_names (e.g. parties as default) to be parties if
        self has no attribute named party_names. Otherwise, parties will be extend to
        the attribute party_names of self.
        
        Args:
            parties (list): a list of objects
            parties_name (str): the name of attribute to store parties

        Example:
        ```python
            >>> a = BasicParty()
            >>> b = BasicParty()
            >>> c = BasicParty()
            >>> a.register_objects([b, c], 'parties')
            >>> a.parties # will be [b,c]
            >>> d = BasicParty()
            >>> a.register_objects([d], 'parties')
            >>> a.parties # will be [b,c,d]
        ```
        """
        if type(parties) is not list:
            raise TypeError("parties should be a list")
        if not hasattr(self, parties_name):
            setattr(self, parties_name, parties)
        else:
            tmp = getattr(self, parties_name)
            if tmp is None: tmp = []
            elif type(tmp) is not list: tmp = list(tmp)
            tmp.extend(parties)
            setattr(self, parties_name, tmp)
        self._object_map.update({p.id:p for p in parties if p.id is not None})

    def communicate_with(self, target_id, package={}):
        r"""
        Send the package to target object according to its id, and receive the response from it

        Args:
            target_id (int): the id of the object to communicate with
            package (dict): the package to be sended to the object
        Returns:
            client_package (dict): the reply from the target object and will be 'None' if losing connection
        """
        return self.gv.communicator.request(self.id, target_id, package)

    def initialize(self, *args, **kwargs):
        r"""API for customizing the initializing process of the object"""
        return

class BasicServer(BasicParty):
    TaskCalculator = flgo.benchmark.base.BasicTaskCalculator
    def __init__(self, option={}):
        super().__init__()
        self.test_data = None
        self.val_data = None
        self.train_data = None
        self.model = None
        self.clients = []
        # basic configuration
        self.task = option['task']
        self.eval_interval = option['eval_interval']
        self.num_parallels = option['num_parallels']
        # server calculator
        self.device = self.gv.apply_for_device() if not option['server_with_cpu'] else torch.device('cpu')
        self.calculator = self.TaskCalculator(self.device, optimizer_name=option['optimizer'])
        # hyper-parameters during training process
        self.num_rounds = option['num_rounds']
        self.num_steps = option['num_steps']
        self.num_epochs = option['num_epochs']
        self.proportion = option['proportion']
        self.decay_rate = option['learning_rate_decay']
        self.lr_scheduler_type = option['lr_scheduler']
        self.learning_rate = option['learning_rate']
        self.sample_option = option['sample']
        self.aggregation_option = option['aggregate']
        # systemic option
        self.tolerance_for_latency = 999999
        # algorithm-dependent parameters
        self.algo_para = {}
        self.current_round = 1
        # all options
        self.option = option
        self.id = -1

    def run(self):
        """
        Running the FL symtem where the global model is trained and evaluated iteratively.
        """
        self.gv.logger.time_start('Total Time Cost')
        if not self._load_checkpoint() and self.eval_interval>0:
            # evaluating initial model performance
            self.gv.logger.info("--------------Initial Evaluation--------------")
            self.gv.logger.time_start('Eval Time Cost')
            self.gv.logger.log_once()
            self.gv.logger.time_end('Eval Time Cost')
        while True:
            if self._if_exit(): break
            self.gv.clock.step()
            # iterate
            updated = self.iterate()
            # using logger to evaluate the model if the model is updated
            if updated is True or updated is None:
                self.gv.logger.info("--------------Round {}--------------".format(self.current_round))
                # check log interval
                if self.gv.logger.check_if_log(self.current_round, self.eval_interval):
                    self.gv.logger.time_start('Eval Time Cost')
                    self.gv.logger.log_once()
                    self.gv.logger.time_end('Eval Time Cost')
                    self._save_checkpoint()
                # check if early stopping
                if self.gv.logger.early_stop(): break
                self.current_round += 1
                # decay learning rate
                self.global_lr_scheduler(self.current_round)
        self.gv.logger.info("=================End==================")
        self.gv.logger.time_end('Total Time Cost')
        # save results as .json file
        self.gv.logger.save_output_as_json()
        return

    def register_exit_condition(self, func: Callable):
        """
        Set exiting condition. Default exiting condition is the number of training rounds reaches a given threshold.
        Args:
            func (Callable): a callable function that takes the server as input and outputs a bool variable to decide whether to stop training
        """
        assert isinstance(func, Callable)
        def new_exit():
            return func(self)
        setattr(self, '_if_exit', new_exit)

    def _if_exit(self):
        """
        Stop federated training when this method returns True. Exiting if the training rounds reaches a given threshold is set as default.
        """
        return self.current_round > self.num_rounds

    def iterate(self):
        """
        The standard iteration of each federated communication round that contains three
        necessary procedure in FL: client selection, communication and model aggregation.

        Returns:
            False if the global model is not updated in this iteration
        """
        # sample clients: MD sampling as default
        self.selected_clients = self.sample()
        # training
        models = self.communicate(self.selected_clients)['model']
        # aggregate: pk = 1/K as default where K=len(selected_clients)
        self.model = self.aggregate(models)
        return len(models) > 0

    @ss.with_clock
    def communicate(self, selected_clients, mtype=0, asynchronous=False):
        """
        The whole simulating communication procedure with the selected clients.
        This part supports for simulating the client dropping out.

        Args:
            selected_clients (list of int): the clients to communicate with
            mtype (anytype): type of message
            asynchronous (bool): asynchronous communciation or synchronous communcation

        Returns:
            :the unpacked response from clients that is created ny self.unpack()
        """
        packages_received_from_clients = []
        received_package_buffer = {}
        communicate_clients = list(set(selected_clients))
        # prepare packages for clients
        for client_id in communicate_clients:
            received_package_buffer[client_id] = None
        # communicate with selected clients
        if self.num_parallels <= 1:
            # computing iteratively
            for client_id in tqdm(communicate_clients, desc="Local Training on {} Clients".format(len(communicate_clients)), leave=False):
                server_pkg = self.pack(client_id, mtype)
                server_pkg['__mtype__'] = mtype
                response_from_client_id = self.communicate_with(self.clients[client_id].id, package=server_pkg)
                packages_received_from_clients.append(response_from_client_id)
        else:
            self.model = self.model.to(torch.device('cpu'))
            try:
                import ray
            except:
                ray = None
            paratype = self.option.get('parallel_type', None)
            if paratype=='r' and ray is not None:
                @ray.remote(num_gpus=min(len(communicate_clients), self.num_parallels)*1.0/len(self.gv.dev_list))
                def wrap_communicate_with(server, client_id, package):
                    return server.communicate_with(client_id, package)
                for client_id in communicate_clients:
                    server_pkg = self.pack(client_id, mtype)
                    server_pkg['__mtype__'] = mtype
                    self.clients[client_id].update_device(self.gv.apply_for_device())
                    res_ref = wrap_communicate_with.remote(self, self.clients[client_id].id, server_pkg)
                    packages_received_from_clients.append(res_ref)
                ready_refs, remaining_refs = ray.wait(packages_received_from_clients, num_returns=len(communicate_clients), timeout=None)
                packages_received_from_clients = ray.get(ready_refs)
            else:
                # computing in parallel with torch.multiprocessing
                if paratype=='t':
                    pool = dmp.Pool(self.num_parallels)
                else:
                    pool = mp.Pool(self.num_parallels)
                for client_id in communicate_clients:
                    server_pkg = self.pack(client_id, mtype)
                    server_pkg['__mtype__'] = mtype
                    self.clients[client_id].update_device(self.gv.apply_for_device())
                    args = (self.clients[client_id].id, server_pkg)
                    packages_received_from_clients.append(pool.apply_async(self.communicate_with, args=args))
                pool.close()
                pool.join()
                packages_received_from_clients = list(map(lambda x: x.get(), packages_received_from_clients))

            self.model = self.model.to(self.device)
            for pkg in packages_received_from_clients:
                for k, v in pkg.items():
                    if hasattr(v, 'to'):
                        try:
                            pkg[k] = v.to(self.device)
                        except:
                            continue
        for i, client_id in enumerate(communicate_clients): received_package_buffer[client_id] = packages_received_from_clients[i]
        packages_received_from_clients = [received_package_buffer[cid] for cid in selected_clients if
                                          received_package_buffer[cid]]
        self.received_clients = selected_clients
        return self.unpack(packages_received_from_clients)

    def communicate_with(self, target_id, package={}):
        r"""Communicate with the object under system simulator that simulates the
        network latency. Send the package to target object according to its id,
        and receive the response from it

        Args:
            target_id (int): the id of the object to communicate with
            package (dict): the package to be sended to the object

        Returns:
            client_package (dict): the reply from the target object and
            will be 'None' if losing connection
        """
        return super(BasicServer, self).communicate_with(target_id, package)

    def pack(self, client_id, mtype=0, *args, **kwargs):
        r"""
        Pack the necessary information for the client's local_movielens_recommendation training.
        Any operations of compression or encryption should be done here.

        Args:
            client_id (int): the id of the client to communicate with
            mtype: the message type

        Returns:
            a dict contains necessary information (e.g. a copy of the global model as default)
        """
        return {
            "model": copy.deepcopy(self.model),
        }

    def unpack(self, packages_received_from_clients):
        r"""
        Unpack the information from the received packages. Return models and losses as default.

        Args:
            packages_received_from_clients (list): a list of packages

        Returns:
            res (dict): collections.defaultdict that contains several lists of the clients' reply
        """
        if len(packages_received_from_clients) == 0: return collections.defaultdict(list)
        res = {pname: [] for pname in packages_received_from_clients[0]}
        for cpkg in packages_received_from_clients:
            for pname, pval in cpkg.items():
                res[pname].append(pval)
        return res

    def global_lr_scheduler(self, current_round):
        r"""
        Control the step size (i.e. learning rate) of local_movielens_recommendation training
        Args:
            current_round (int): the current communication round
        """
        if self.lr_scheduler_type == '-1' or self.lr_scheduler_type.lower()=='constant':
            return
        elif self.lr_scheduler_type == '0' or self.lr_scheduler_type.lower().startswith('step'):
            """eta_{round+1} = DecayRate * eta_{round}"""
            step_size = self.lr_scheduler_type.split('_')
            step_size = max(int(step_size[-1]), 1) if len(step_size)>1 else 1
            if current_round>0 and current_round%step_size==0:
                self.learning_rate *= self.decay_rate
                for c in self.clients:
                    c.set_learning_rate(self.learning_rate)
        elif self.lr_scheduler_type == '1' or self.lr_scheduler_type=='divide_by_round':
            """eta_{round+1} = eta_0/(round+1)"""
            self.learning_rate = self.option['learning_rate'] * 1.0 / (current_round + 1)
            for c in self.clients:
                c.set_learning_rate(self.learning_rate)
        elif self.lr_scheduler_type.startswith('multistep'):
            """eta_{step_round} = DecayRate * eta_{current}"""
            rounds = self.lr_scheduler_type.split('_')[1:]
            if len(rounds)==0: return
            rounds = [int(r) for r in rounds]
            if current_round in rounds:
                self.learning_rate *= self.decay_rate
                for c in self.clients:
                    c.set_learning_rate(self.learning_rate)
        elif self.lr_scheduler_type.startswith('ratiostep'):
            """eta_{k*int(num_rounds*ratio)} = DecayRate * eta_{current}"""
            ratio = min(float(self.lr_scheduler_type.split('_')[-1]), 1.0)
            step_round = max(int(self.num_rounds*ratio), 1)
            if current_round%step_round:
                self.learning_rate *= self.decay_rate
                for c in self.clients:
                    c.set_learning_rate(self.learning_rate)

    def sample(self):
        r"""
        Sample the clients. There are three types of sampling manners:
        full sample, uniform sample without replacement, and MDSample
        with replacement. Particularly, if 'available' is in self.sample_option,
        the server will only sample from currently available clients.

        Returns:
            a list of the ids of the selected clients

        Example:
        ```python
            >>> selected_clients=self.sample()
            >>> selected_clients
            >>> # The selected_clients is a list of clients' ids
        ```
        """
        all_clients = self.available_clients if 'available' in self.sample_option else [cid for cid in
                                                                                        range(self.num_clients)]
        clients_per_round = max(min(int(self.num_clients*self.proportion),len(all_clients)), 1)
        # full sampling with unlimited communication resources of the server
        if 'full' in self.sample_option:
            return all_clients
        # sample clients
        elif 'uniform' in self.sample_option:
            # original sample proposed by fedavg
            selected_clients = list(
                np.random.choice(all_clients, clients_per_round, replace=False)) if len(
                all_clients) > 0 else []
        elif 'md' in self.sample_option:
            # the default setting that is introduced by FedProx, where the clients are sampled with the probability in proportion to their local_movielens_recommendation data sizes
            local_data_vols = [self.clients[cid].datavol for cid in all_clients]
            total_data_vol = sum(local_data_vols)
            p = np.array(local_data_vols) / total_data_vol
            selected_clients = list(np.random.choice(all_clients, clients_per_round, replace=True, p=p)) if len(
                all_clients) > 0 else []
        return selected_clients

    def aggregate(self, models: list, *args, **kwargs):
        r"""
        Aggregate the locally trained models into the new one. The aggregation
        will be according to self.aggregate_option where

        pk = nk/n where n=self.data_vol
        K = |S_t|
        N = |S|
        -------------------------------------------------------------------------------------------------------------------------
         weighted_scale                 |uniform (default)          |weighted_com (original fedavg)   |other
        ==========================================================================================================================
        N/K * Σpk * model_k             |1/K * Σmodel_k             |(1-Σpk) * w_old + Σpk * model_k  |Σ(pk/Σpk) * model_k


        Args:
            models (list): a list of local_movielens_recommendation models

        Returns:
            the aggregated model

        Example:
        ```python
            >>> models = [m1, m2] # m1, m2 are models with the same architecture
            >>> m_new = self.aggregate(models)
        ```
        """
        if len(models) == 0: return self.model
        nan_exists = [m.has_nan() for m in models]
        if any(nan_exists):
            if all(nan_exists): raise ValueError("All the received local models have parameters of nan value.")
            self.gv.logger.info('Warning("There exists nan-value in local models, which will be automatically removed from the aggregatino list.")')
            new_models = []
            received_clients = []
            for ni, mi, cid in zip(nan_exists, models, self.received_clients):
                if ni: continue
                new_models.append(mi)
                received_clients.append(cid)
            self.received_clients = received_clients
            models = new_models
        local_data_vols = [c.datavol for c in self.clients]
        total_data_vol = sum(local_data_vols)
        if self.aggregation_option == 'weighted_scale':
            p = [1.0 * local_data_vols[cid] / total_data_vol for cid in self.received_clients]
            K = len(models)
            N = self.num_clients
            return fmodule._model_sum([model_k * pk for model_k, pk in zip(models, p)]) * N / K
        elif self.aggregation_option == 'uniform':
            return fmodule._model_average(models)
        elif self.aggregation_option == 'weighted_com':
            p = [1.0 * local_data_vols[cid] / total_data_vol for cid in self.received_clients]
            w = fmodule._model_sum([model_k * pk for model_k, pk in zip(models, p)])
            return (1.0 - sum(p)) * self.model + w
        else:
            p = [1.0 * local_data_vols[cid] / total_data_vol for cid in self.received_clients]
            sump = sum(p)
            p = [pk / sump for pk in p]
            return fmodule._model_sum([model_k * pk for model_k, pk in zip(models, p)])

    def global_test(self, model=None, flag:str='val'):
        r"""
        Collect local_movielens_recommendation testing result of all the clients.

        Args:
            model (flgo.utils.fmodule.FModule|torch.nn.Module): the model to be sevaluated
            flag (str): choose the data to evaluate the model

        Returns:
            metrics (dict): a dict contains key-value pairs like (metric_name,
            the lists of metric results of the clients)
        """
        if model is None: model=self.model
        all_metrics = collections.defaultdict(list)
        for c in self.clients:
            client_metrics = c.test(model, flag)
            for met_name, met_val in client_metrics.items():
                all_metrics[met_name].append(met_val)
        return all_metrics

    def test(self, model=None, flag:str='test'):
        r"""
        Evaluate the model on the test dataset owned by the server.

        Args:
            model (flgo.utils.fmodule.FModule): the model need to be evaluated
            flag (str): choose the data to evaluate the model

        Returns:
            metrics (dict): the dict contains the evaluating results
        """
        if model is None: model = self.model
        dataset = getattr(self, flag+'_data') if hasattr(self, flag+'_data') else None
        if dataset is None:
            return {}
        else:
            return self.calculator.test(model, dataset, batch_size=min(self.option['test_batch_size'], len(dataset)),
                                        num_workers=self.option['num_workers'], pin_memory=self.option['pin_memory'])

    def init_algo_para(self, algo_para: dict):
        """
        Initialize the algorithm-dependent hyper-parameters for the server and all the clients.

        Args:
            algo_paras (dict): the dict that defines the hyper-parameters (i.e. name, value and type) for the algorithm.

        Example:
        ```python
            >>> # s is an instance of Server and s.clients are instances of Client
            >>> s.u # will raise error
            >>> [c.u for c in s.clients] # will raise errors too
            >>> s.init_algo_para({'u': 0.1})
            >>> s.u # will be 0.1
            >>> [c.u for c in s.clients] # will be [0.1, 0.1,..., 0.1]
        ```
        Note:
            Once `option['algo_para']` is not `None`, the value of the pre-defined hyperparameters will be replaced by the list of values in `option['algo_para']`,
            which requires the length of `option['algo_para']` is equal to the length of `algo_paras`
        """
        self.algo_para = algo_para
        if len(self.algo_para) == 0: return
        # initialize algorithm-dependent hyperparameters from the input options
        if self.option['algo_para'] is not None:
            # assert len(self.algo_para) == len(self.option['algo_para'])
            keys = list(self.algo_para.keys())
            for i, pv in enumerate(self.option['algo_para']):
                if i == len(self.option['algo_para']): break
                para_name = keys[i]
                try:
                    self.algo_para[para_name] = type(self.algo_para[para_name])(pv)
                except:
                    self.algo_para[para_name] = pv
        # register the algorithm-dependent hyperparameters as the attributes of the server and all the clients
        for para_name, value in self.algo_para.items():
            self.__setattr__(para_name, value)
            for p in self._object_map.values():
                p.__setattr__(para_name, value)
        return

    def get_tolerance_for_latency(self):
        r"""
        Get the tolerance for latency of waiting for clients' responses

        Returns:
            a int number (i.e. self.tolerance_for_latency)
        """
        return self.tolerance_for_latency

    def set_tolerance_for_latency(self, tolerance:int):
        r"""
        Set the tolerance for latency of waiting for clients' responses

        Args:
            tolerance (int): the amounts of virtual time units
        """
        self.tolerance_for_latency = tolerance

    def wait_time(self, t=1):
        r"""
        Wait for the time of the virtual clock to pass t units
        """
        self.gv.clock.step(t)
        return

    @property
    def available_clients(self):
        """
        Return all the available clients at the current round.

        Returns:
            a list of indices of currently available clients
        """
        return [cid for cid in range(self.num_clients) if self.clients[cid].is_idle()]

    def register_clients(self, clients):
        """
        Regiser clients to self.clients, and update related attributes (e.g. self.num_clients)
        
        Args:
            clients (list): a list of objects
        """
        self.register_objects(clients, 'clients')
        self.num_clients = len(clients)
        for cid, c in enumerate(self.clients):
            c.client_id = cid
        for c in self.clients: c.register_server(self)
        self.clients_per_round = max(int(self.num_clients * self.proportion), 1)
        self.selected_clients = []
        self.dropped_clients = []

    def _save_checkpoint(self):
        checkpoint = self.option.get('save_checkpoint', '')
        if checkpoint!='':
            cpt = self.save_checkpoint()
            cpt_path = os.path.join(self.option['task'], 'checkpoint', checkpoint,)
            if not os.path.exists(cpt_path): os.makedirs(cpt_path)
            cpt_path = os.path.join(cpt_path, self.gv.logger.get_output_name(f'.{self.current_round}'))
            torch.save(cpt, cpt_path)

    def save_checkpoint(self):
        cpt = {
            'round': self.current_round,
            'learning_rate': self.learning_rate,
            'model_state_dict': self.model.state_dict(),
            'early_stop_option': {
                '_es_best_score': self.gv.logger._es_best_score,
                '_es_best_round': self.gv.logger._es_best_round,
                '_es_patience': self.gv.logger._es_patience,
            },
            'output': self.gv.logger.output,
            'time': self.gv.clock.current_time,
        }
        return cpt

    def load_checkpoint(self, cpt):
        md = cpt.get('model_state_dict', None)
        round = cpt.get('round', None)
        output = cpt.get('output', None)
        early_stop_option = cpt.get('early_stop_option', None)
        time = cpt.get('time', None)
        learning_rate = cpt.get('learning_rate', None)
        if md is not None: self.model.load_state_dict(md)
        if round is not None: self.current_round = round + 1
        if output is not None: self.gv.logger.output = output
        if time is not None: self.gv.clock.set_time(time)
        if learning_rate is not None: self.learning_rate = learning_rate
        if early_stop_option is not None:
            self.gv.logger._es_best_score = early_stop_option['_es_best_score']
            self.gv.logger._es_best_round = early_stop_option['_es_best_round']
            self.gv.logger._es_patience = early_stop_option['_es_patience']


    def _load_checkpoint(self):
        checkpoint = self.option.get('load_checkpoint', '')
        if checkpoint!='':
            try:
                cpt_name = self.gv.logger.get_output_name('')
                cpt_path = os.path.join(self.option['task'], 'checkpoint', checkpoint)
                cpts = os.listdir(cpt_path)
                cpts = [p for p in cpts if cpt_name in p]
                cpt_round = max([int(p.split('.')[-1]) for p in cpts])
                cpt_path = os.path.join(cpt_path, '.'.join([cpt_name, str(cpt_round)]))
                self.gv.logger.info(f'Loading checkpoint {cpt_name} at round {cpt_round}...')
                cpt = torch.load(cpt_path)
                self.load_checkpoint(cpt)
                return True
            except Exception as e:
                self.gv.logger.info(f"Failed to load checkpoint due to {e}")
        return False

class BasicClient(BasicParty):
    TaskCalculator = flgo.benchmark.base.BasicTaskCalculator
    def __init__(self, option={}):
        super().__init__()
        self.id = None
        # create local_movielens_recommendation dataset
        self.data_loader = None
        self.test_data = None
        self.val_data = None
        self.train_data = None
        self.model = None
        # local_movielens_recommendation calculator
        self.device = self.gv.apply_for_device()
        self.calculator = self.TaskCalculator(self.device, option['optimizer'])
        self._train_loader = None
        # hyper-parameters for training
        self.optimizer_name = option['optimizer']
        self.learning_rate = option['learning_rate']
        self.momentum = option['momentum']
        self.weight_decay = option['weight_decay']
        self.batch_size = option['batch_size']
        self.num_steps = option['num_steps']
        self.num_epochs = option['num_epochs']
        self.clip_grad = option['clip_grad']
        self.model = None
        self.test_batch_size = option['test_batch_size']
        self.loader_num_workers = option['num_workers']
        self.current_steps = 0
        # system setting
        self._effective_num_steps = self.num_steps
        self._latency = 0
        # server
        self.server = None
        # actions of different message type
        self.option = option
        self.actions = {0: self.reply}
        self.default_action = self.reply

    @fmodule.with_multi_gpus
    def train(self, model):
        r"""
        Standard local_movielens_recommendation training procedure. Train the transmitted model with
        local_movielens_recommendation training dataset.

        Args:
            model (FModule): the global model
        """
        model.train()
        optimizer = self.calculator.get_optimizer(model, lr=self.learning_rate, weight_decay=self.weight_decay,
                                                  momentum=self.momentum)
        for iter in range(self.num_steps):
            # get a batch of data
            batch_data = self.get_batch_data()
            model.zero_grad()
            # calculate the loss of the model on batched dataset through task-specified calculator
            loss = self.calculator.compute_loss(model, batch_data)['loss']
            loss.backward()
            if self.clip_grad>0:torch.nn.utils.clip_grad_norm_(parameters=model.parameters(), max_norm=self.clip_grad)
            optimizer.step()
        return

    @fmodule.with_multi_gpus
    def test(self, model, flag='val'):
        r"""
        Evaluate the model on the dataset owned by the client

        Args:
            model (flgo.utils.fmodule.FModule): the model need to be evaluated
            flag (str): choose the data to evaluate the model

        Returns:
            metric (dict): the evaluating results (e.g. metric = {'loss':1.02})
        """
        dataset = getattr(self, flag + '_data') if hasattr(self, flag + '_data') else None
        if dataset is None: return {}
        return self.calculator.test(model, dataset, min(self.test_batch_size, len(dataset)), self.option['num_workers'])

    def unpack(self, received_pkg):
        r"""
        Unpack the package received from the server

        Args:
            received_pkg (dict): a dict contains the global model as default

        Returns:
            the unpacked information
        """
        # unpack the received package
        return received_pkg['model']

    def reply(self, svr_pkg):
        r"""
        Reply a package to the server. The whole local_movielens_recommendation procedure should be defined here.
        The standard form consists of three procedure: unpacking the
        server_package to obtain the global model, training the global model,
        and finally packing the updated model into client_package.

        Args:
            svr_pkg (dict): the package received from the server

        Returns:
            client_pkg (dict): the package to be send to the server
        """
        model = self.unpack(svr_pkg)
        self.train(model)
        cpkg = self.pack(model)
        return cpkg

    def pack(self, model, *args, **kwargs):
        r"""
        Packing the package to be send to the server. The operations of compression
        of encryption of the package should be done here.

        Args:
            model: the locally trained model

        Returns:
            package: a dict that contains the necessary information for the server
        """
        return {
            "model": model,
        }

    def is_idle(self):
        r"""
        Check if the client is available to participate training.

        Returns:
            True if the client is available according to the active_rate else False
        """
        return self.gv.simulator.client_states[self.id] == 'idle'

    def is_dropped(self):
        r"""
        Check if the client drops out during communicating.

        Returns:
            True if the client was being dropped
        """
        return self.gv.simulator.client_states[self.id] == 'dropped'

    def is_working(self):
        r"""
        Check if the client is training the model.

        Returns:
            True if the client is working
        """

        return self.gv.simulator.client_states[self.id] == 'working'

    def train_loss(self, model):
        r"""
        Get the loss value of the model on local_movielens_recommendation training data

        Args:
            model (flgo.utils.fmodule.FModule|torch.nn.Module): model

        Returns:
            the training loss of model on self's training data
        """
        return self.test(model, 'train')['loss']

    def val_loss(self, model):
        r"""
        Get the loss value of the model on local_movielens_recommendation validating data

        Args:
            model (flgo.utils.fmodule.FModule|torch.nn.Module): model

        Returns:
            the validation loss of model on self's validation data
        """
        return self.test(model)['loss']

    def register_server(self, server=None):
        r"""
        Register the server to self.server
        """
        self.register_objects([server], 'server_list')
        if server is not None:
            self.server = server

    def set_local_epochs(self, epochs=None):
        r"""
        Set local_movielens_recommendation training epochs
        """
        if epochs is None: return
        self.epochs = epochs
        self.num_steps = self.epochs * math.ceil(len(self.train_data) / self.batch_size)
        return

    def set_batch_size(self, batch_size=None):
        r"""
        Set local_movielens_recommendation training batch size

        Args:
            batch_size (int): the training batch size
        """
        if batch_size is None: return
        self.batch_size = batch_size

    def set_learning_rate(self, lr=None):
        """
        Set the learning rate of local_movielens_recommendation training
        Args:
            lr (float): a real number
        """
        self.learning_rate = lr if lr else self.learning_rate

    def get_time_response(self):
        """
        Get the latency amount of the client

        Returns:
            self.latency_amount if client not dropping out
        """
        return np.inf if self.dropped else self.time_response

    def get_batch_data(self):
        """
        Get the batch of training data
        Returns:
            a batch of data
        """
        if self._train_loader is None:
            self._train_loader = self.calculator.get_dataloader(self.train_data, batch_size=self.batch_size,
                                                                   num_workers=self.loader_num_workers,
                                                                   pin_memory=self.option['pin_memory'], drop_last=self.option.get('drop_last', False))
        try:
            batch_data = next(self.data_loader)
        except Exception as e:
            self.data_loader = iter(self._train_loader)
            batch_data = next(self.data_loader)
        # clear local_movielens_recommendation DataLoader when finishing local_movielens_recommendation training
        self.current_steps = (self.current_steps + 1) % self.num_steps
        if self.current_steps == 0:
            self.data_loader = None
            self._train_loader = None
        return batch_data

    def update_device(self, dev):
        """
        Update running-time GPU device to dev

        Args:
            dev (torch.device): target dev
        """
        self.device = dev
        self.calculator = self.gv.TaskCalculator(dev, self.calculator.optimizer_name)



import h5py
import torch_geometric.data
from flgo.benchmark.toolkits import BasicTaskGenerator, BasicTaskCalculator, BasicTaskPipe
import os.path
import torch
from torch.utils.data import DataLoader
import os
import zipfile
from urllib.request import urlopen
import urllib.request
import numpy as np
import flgo
import pandas as pd
import scipy.sparse as sp
try:
    import ujson as json
except:
    import json

def download_from_url(url= None, filepath = '.'):
    """Download dataset from url to filepath."""
    if url: urllib.request.urlretrieve(url, filepath)
    return filepath

class TaskGenerator(BasicTaskGenerator):
    _DATASET_INFO = {
        'url':"https://raw.githubusercontent.com/fmonti/mgcnn/master/Data/douban/training_test_dataset.mat",
        'link_type':'trust',
        'file': 'training_test_dataset.mat',
    }
    #
    def __init__(self, rawdata_path=os.path.join(flgo.benchmark.data_root, 'DOUBAN')):
        super(TaskGenerator, self).__init__(benchmark='social_splitted_douban', rawdata_path=rawdata_path)
        self.scene='vertical'
        for key,value in self._DATASET_INFO.items(): setattr(self, key, value)
        self.num_users = 0

    def partition(self, *args, **kwargs):
        pass

    def get_task_name(self):
        return 'social_splitted_douban'

    def load_data(self, *args, **kwargs):
        # download zip file
        filepath = os.path.join(self.rawdata_path, self.file)
        if not os.path.exists(filepath):
            if not os.path.exists(self.rawdata_path):
                os.makedirs(self.rawdata_path)
            _ = download_from_url(self.url, filepath)
        # read data into memory
        db = h5py.File(filepath, 'r')
        fields = ['M', 'Otraining', 'Otest', 'W_users']
        outputs = []
        for name_field in fields:
            ds = db[name_field]
            try:
                if 'ir' in ds.keys():
                    data = np.asarray(ds['data'])
                    ir = np.asarray(ds['ir'])
                    jc = np.asarray(ds['jc'])
                    out = sp.csc_matrix((data, ir, jc)).astype(np.float32)
            except AttributeError:
                # Transpose in case is a dense matrix because of the row- vs column- major ordering between python and matlab
                out = np.asarray(ds).astype(np.float32).T
            outputs.append(out)
        M, Otraining, Otest, Wrow = outputs
        self.num_users=3000
        self.num_items = 3000
        train_idxs = np.where(Otraining>0)
        train_ratings = M[train_idxs]
        test_idxs = np.where(Otest>0)
        test_ratings = M[test_idxs]
        train_dict = {'userID':train_idxs[0].tolist(), 'movieID':train_idxs[1].tolist(), 'Rating':train_ratings.tolist()}
        test_dict = {'userID':test_idxs[0].tolist(), 'movieID':test_idxs[1].tolist(), 'Rating':test_ratings.tolist()}
        self.ratings_train = pd.DataFrame(train_dict)
        self.ratings_test = pd.DataFrame(test_dict)
        link_idxs = np.where(Wrow>0)
        link_values = Wrow[link_idxs]
        link_dict = {'trustorID':link_idxs[0].tolist(), 'trusteeID':link_idxs[1].tolist(), 'Value':link_values}
        self.links = pd.DataFrame(link_dict)
        self.users = list(range(self.num_users))
        self.items = list(range(self.num_items))
        return

class TaskPipe(BasicTaskPipe):
    class TaskDataset(torch.utils.data.Dataset):
        def __init__(self, data):
            self.user_ids = torch.LongTensor([d[0] for d in data])
            self.movie_ids = torch.LongTensor([d[1] for d in data])
            self.ratings = torch.FloatTensor([d[2] for d in data])

        def __getitem__(self, item):
            return self.user_ids[item], self.movie_ids[item], self.ratings[item]

        def __len__(self):
            return len(self.user_ids)

    def save_task(self, generator):
        party_names = ['Advertiser', 'MeidaCompany']
        feddata = {'party_names': party_names, 'rawdata_path': generator.rawdata_path}
        feddata['Advertiser'] = {
            'data':{
                'userID':generator.users,
                'itemID':generator.items,
                'ratings_train': generator.ratings_train.values.tolist(),
                'ratings_test': generator.ratings_test.values.tolist(),
                'with_label':True
            }
        }
        feddata['MeidaCompany'] = {
            'data': {
                'userID': generator.users,
                'links': generator.links.values.tolist(),
                'with_label':False
            }
        }
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)

    def load_data(self, running_time_option, *args, **kwargs):
        task_data = {}
        for pid, party_name in enumerate(self.feddata['party_names']):
            pdata = self.feddata[party_name]['data']
            if party_name=='Advertiser':
                users = pdata['userID']
                test_data = pdata['ratings_test']
                ratings_train = pdata['ratings_train']
                tmp = pd.DataFrame(ratings_train)
                train_data = []
                val_data = []
                val_ratio = running_time_option['train_holdout']
                for uid in users:
                    udata = tmp[tmp[0]==uid].values.tolist()
                    train_size = max(int(len(udata)*(1-val_ratio)), 1)
                    utrain = udata[:train_size]
                    uvalid = udata[train_size:]
                    train_data.extend(utrain)
                    val_data.extend(uvalid)
                train_data = self.TaskDataset(train_data)
                val_data = self.TaskDataset(val_data) if len(val_data)>0 else None
                test_data = self.TaskDataset(test_data)
                task_data[party_name] = {'userID': pdata['userID'], 'itemID':pdata['itemID'], 'train':train_data, 'val':val_data, 'test':test_data}
            else:
                social_links = pdata['links']
                edge_index = torch.tensor([[l[0] for l in social_links], [l[1] for l in social_links]], dtype=torch.int64)
                social_data = torch_geometric.data.Data(edge_index=edge_index)
                task_data[party_name] = {'userID':pdata['userID'],'social':social_data}
        return task_data

class TaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super().__init__(device, optimizer_name)
        self.criterion = torch.nn.MSELoss()
        self.DataLoader = DataLoader

    def to_device(self, data):
        return data[0].to(self.device), data[1].to(self.device), data[2].to(self.device)
#
    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False, drop_last=False, *args, **kwargs):
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, drop_last=drop_last)

    def compute_loss(self, model, data, *args, **kwargs):
        tdata = self.to_device(data)
        model = model.to(self.device)
        output = model(tdata)
        loss =  self.criterion(output, torch.tensor(tdata[-1], dtype=torch.float))
        return {'loss':loss}

    def test(self, model, dataset, batch_size=64, num_workers = 0, *args, **kwargs):
        model.to(self.device)
        with torch.no_grad():
            model.eval()
            if batch_size == -1: batch_size = len(dataset)
            data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers)
            mse_loss = 0
            mae_loss = 0
            for batch_id, batch_data in enumerate(data_loader):
                batch_data = self.to_device(batch_data)
                outputs = model(batch_data)
                batch_mae = torch.abs(outputs-batch_data[-1]).sum().item()
                batch_mse = torch.nn.MSELoss()(outputs, batch_data[-1]).item()
                mse_loss += batch_mse * len(batch_data[-1])
                mae_loss += batch_mae
            mae = mae_loss/len(dataset)
            rmse = torch.sqrt(torch.tensor(mse_loss/len(dataset)))
        return {'mae': mae, 'rmse': rmse.item()}


import flgo.benchmark.toolkits.partition
import flgo.benchmark.heart_disease_classification.model.lr as lr

default_partitioner = flgo.benchmark.toolkits.partition.IIDPartitioner
default_partition_para = {'num_clients':100}
default_model = lr


import multiprocessing
import os
import time
import uuid
import flgo.utils.fflow as fuf
import torch
import numpy as np
import torch.utils.data as tud
from multiprocessing.managers import SharedMemoryManager
from multiprocessing import shared_memory
from shared_memory_dict import SharedMemoryDict
import sys

def get_dict_size(obj, seen=None):
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    seen.add(obj_id)
    size = sys.getsizeof(obj)
    if isinstance(obj, dict):
        for key, value in obj.items():
            size += sys.getsizeof(key) + get_dict_size(value, seen)
    elif isinstance(obj, (list, tuple, set)):
        for item in obj:
            size += get_dict_size(item, seen)
    return size

TYPE_CANDIDATES = ['int', 'float', 'str', 'ndarray', 'Tensor', 'list', 'tuple', 'dict']

class TmpDataset(tud.Dataset):
    def __init__(self, sharable_data, types):
        super().__init__()
        data = []
        for i in range(len(types)):
            if types[i] == 'ndarray':
                data.append(sharable_data[i])
            elif types[i] == 'Tensor':
                data.append(torch.from_numpy(sharable_data[i]))
            elif types[i] in ['int', 'float', 'str']:
                data.append(sharable_data[i].tolist())
        self.data = data

    def __len__(self):
        return len(self.data[0])

    def __getitem__(self, i):
        return tuple(d[i] for d in self.data)

def dataset2sharable(dataset):
    first_item = dataset[0]
    item_size = len(first_item)
    if not isinstance(first_item, tuple): first_item = tuple(first_item)
    element_types = [type(ei).__name__ if type(ei) not in TYPE_CANDIDATES else 'unknown' for ei in first_item]
    dataset_size = len(dataset)
    res = [[] for _ in range(item_size)]
    for i in range(dataset_size):
        for j in range(item_size):
            res[j].append(dataset[i][j])
    for j in range(item_size):
        if type(element_types[j]).__name__ in ['int', 'float', 'str']:
            res[j] = np.array(res[j])
        elif type(element_types[j]).__name__ == 'ndarray':
            res[j] = np.stack(res[j])
        elif type(element_types[j]).__name__ == 'Tensor':
            res[j] = torch.stack(res[j]).numpy()
        else:
            raise ValueError("Unsupported type")
    return res, element_types

def sharable2dataset(res, element_types):
    return TmpDataset(res, element_types)

def create_meta_for_dataset(sharable_data, name):
    shm_name_list = []
    shm_dtype_list = []
    shm_shape_list = []
    for i, sdi in enumerate(sharable_data):
        shm_name = name+f"{uuid.uuid4()}"
        shm_name_list.append(shm_name)
        shm_dtype_list.append(sdi.dtype.name)
        shm_shape_list.append(sdi.shape)
        dst = np.memmap(shm_name, mode='w+', dtype=sdi.dtype, shape=sdi.shape)
        np.copyto(dst, sdi)
        # d_size = sdi.dtype.itemsize * np.prod(sdi.shape)
        # shm = shared_memory.SharedMemory(create=True, name=shm_name, size=d_size)
        # dst = np.ndarray(shape=sdi.shape, dtype=sdi.dtype, buffer=shm.buf)
        # dst[:] = sdi[:]
    return shm_name_list, shm_dtype_list, shm_shape_list

def create_meta_for_task(task_data):
    task_meta = {}
    for party in task_data:
        task_meta[party] = {}
        for data_name in task_data[party]:
            data = task_data[party][data_name]
            if data is None: continue
            sharable_data, etypes = dataset2sharable(data)
            shm_name_list, shm_type_list, shm_shape_list= create_meta_for_dataset(sharable_data, "_".join([party, data_name]))
            task_meta[party][data_name] = {
                "name": shm_name_list,
                "dtype": shm_type_list,
                "shape": shm_shape_list,
                'element_type': etypes,
            }
    return task_meta

def load_dataset_from_meta(name, dtype, shape, element_type):
    sharable_data = []
    target_element_types = []
    for name_i, dtype_i, shape_i, element_type_i in zip(name, dtype, shape, element_type):
        np_array = np.memmap(name_i, mode='r+', dtype=dtype_i, shape=shape_i)
        sharable_data.append(np_array)
        target_element_types.append(element_type_i)
    return sharable2dataset(sharable_data, target_element_types)

def load_taskdata_from_meta(task_meta):
    task_data = {}
    for party in task_meta:
        task_data[party] = {}
        for data_name in task_meta[party]:
            task_data[party][data_name] = load_dataset_from_meta(**task_meta[party][data_name])
    return task_data

def worker(task_meta):
    pid = os.getpid()
    task_data = load_taskdata_from_meta(task_meta)
    selected_party = 'server'
    dataset = task_data[selected_party]['test']
    dataloader = tud.DataLoader(dataset, batch_size=2)
    for batch_id, batch_data in enumerate(dataloader):
        x, y = batch_data
        print(f"Process {pid} Xid = {id(x.numpy())} {pid} Y = {id(y.numpy())}")
        if batch_id>=3: break
    return True

if __name__=='__main__':
    NUM_WORKERS = 3
    task = './my_task'
    task_data = fuf.load_task_data(task)
    task_meta = create_meta_for_task(task_data)
    pros = []
    for i in range(NUM_WORKERS):
        p = multiprocessing.Process(target=worker,  args=(task_meta, ))
        pros.append(p)
    for p in pros: p.start()
    for p in pros: p.join()



import os
from flgo.benchmark.toolkits.cv.classification import GeneralCalculator, FromDatasetPipe, FromDatasetGenerator
from .config import train_data
from torch.utils.data import Subset, ConcatDataset
import torch
try:
    import ujson as json
except:
    import json
try:
    from .config import test_data
except:
    test_data = None
try:
    from .config import val_data
except:
    val_data = None

class TaskGenerator(FromDatasetGenerator):
    def __init__(self):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1],
                                            train_data=train_data, val_data=val_data, test_data=test_data)

class TaskPipe(FromDatasetPipe):
    TaskDataset = Subset
    def __init__(self, task_path):
        super(FromDatasetPipe, self).__init__(task_path, train_data=train_data, val_data=val_data, test_data=test_data)
        if os.path.exists(os.path.join(self.task_path, 'info')):
            with open(os.path.join(self.task_path, 'info'), 'r') as inf:
                self.info = json.load(inf)

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid],}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def split_dataset(self, dataset, p=0.0):
        if p == 0: return dataset, None
        s1 = int(len(dataset) * p)
        s2 = len(dataset) - s1
        if s1==0:
            return dataset, None
        elif s2==0:
            return None, dataset
        else:
            return torch.utils.data.random_split(dataset, [s2, s1])

    def load_data(self, running_time_option) -> dict:
        train_data = self.train_data
        test_data = self.test_data
        val_data = self.val_data
        # rearrange data for server
        if val_data is None and test_data is not None:
            server_data_val, server_data_test = self.split_dataset(test_data, 1.0-running_time_option['test_holdout'])
        else:
            server_data_test = test_data
            server_data_val = val_data
        task_data = {'server': {'test': server_data_test, 'val': server_data_val}}
        # rearrange data for clients
        partitioner_name = self.info.get('partitioner', None)
        if partitioner_name is not None and 'Concat' in partitioner_name:
            assert hasattr(train_data, 'datasets')
            all_datasets = train_data.datasets
            cdata_trains = []
            cdata_vals = []
            cdata_tests = []
            for cid, cname in enumerate(self.feddata['client_names']):
                dataset_ids = self.feddata[cname]['data']
                cdataset = [all_datasets[did] for did in dataset_ids]
                for cdata_i in cdataset:
                    cdata_train_i, cdata_val_i = self.split_dataset(cdata_i, running_time_option['train_holdout'])
                    if cdata_val_i is not None:
                        cdata_val_i, cdata_test_i = self.split_dataset(cdata_val_i, running_time_option['local_test_ratio'])
                    else:
                        cdata_test_i = None
                    if cdata_train_i is not None: cdata_trains.append(cdata_train_i)
                    if cdata_val_i is not None: cdata_vals.append(cdata_val_i)
                    if cdata_test_i is not None: cdata_tests.append(cdata_test_i)
                cdata_train = ConcatDataset(cdata_trains) if len(cdata_trains)>0 else None
                cdata_val = ConcatDataset(cdata_vals) if len(cdata_vals)>0 else None
                cdata_test = ConcatDataset(cdata_tests) if len(cdata_tests)>0 else None
                task_data[cname] = {'train':cdata_train, 'val':cdata_val, 'test': cdata_test}
        else:
            for cid, cname in enumerate(self.feddata['client_names']):
                cdata = self.TaskDataset(train_data, self.feddata[cname]['data'])
                cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
                if running_time_option['train_holdout']>0 and running_time_option['local_test']:
                    cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
                else:
                    cdata_test = None
                task_data[cname] = {'train':cdata_train, 'val':cdata_val, 'test': cdata_test}
        return task_data

TaskCalculator = GeneralCalculator


import collections
import numpy as np
import torchvision
import flgo.utils.fmodule as fmodule
import torch
from tqdm import tqdm
from sklearn.metrics import auc

class Model(fmodule.FModule):
    CLASSES = [str(i) for i in range(91)]
    def __init__(self):
        super(Model, self).__init__()
        self.model = torchvision.models.detection.fasterrcnn_resnet50_fpn(num_classes=91)

    def transform(self, images, targets, device=None):
        if device is not None:
            return list(image.to(device) for image in images), [{k: v.to(device) for k, v in t.items()} for t in targets]
        return list(image for image in images), [{k: v for k, v in t.items()} for t in targets]

    def organize_output(self, output):
        if self.model.training:
            loss_total = sum(loss for loss in output.values())
            return {'loss':loss_total}
        else:
            return output

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def evaluate(self, dataloader=None):
        if dataloader is None or len(dataloader)==0: return {}
        losses = []
        with torch.no_grad():
            self.model.train()
            for images, targets in tqdm(dataloader):
                images, targets = self.transform(images, targets, device=self.get_device())
                output = self.model(images, targets)
                output = self.organize_output(output)
                losses.append(output['loss']*len(images))
        self.model.eval()
        metric_dict_by_class = [collections.defaultdict(list) for _ in self.CLASSES]
        for images, targets in tqdm(dataloader):
            images, _ = self.transform(images, targets, device=self.get_device())
            output = self.model(images)
            predictions = self.organize_output(output)
            # targets = [
            #     {'boxes': [torch.LongTensor(
            #         [int(ob['bndbox']['xmin']), int(ob['bndbox']['ymin']), int(ob['bndbox']['xmax']),
            #          int(ob['bndbox']['ymax'])]) for ob in t['annotation']['object']],
            #      'labels': [self.CLASSES.index(ob['name']) for ob in t['annotation']['object']]
            #      } for t in targets]
            targets = [
                {
                    'boxes': t['boxes'],
                    'labels': t['labels'],
                } for t in targets
            ]
            for label, class_name in enumerate(self.CLASSES):
                precision, recall, ap, num_class_obj = precision_recall_ap(targets, predictions, iou_thresh=0.5,
                                                                           class_label=label)
                if num_class_obj > 0:
                    mean_ap = mean_average_precision(targets, predictions, class_label=label)
                    acc = accuracy(targets, predictions, class_label=label)
                    red = redundancy(targets, predictions, class_label=label)
                    metric_dict_by_class[label]['num_objs'].append(num_class_obj)
                    metric_dict_by_class[label]['precision'].append(precision)
                    metric_dict_by_class[label]['recall'].append(recall)
                    metric_dict_by_class[label]['ap'].append(ap)
                    metric_dict_by_class[label]['mAP'].append(mean_ap)
                    metric_dict_by_class[label]['acc'].append(acc)
                    metric_dict_by_class[label]['red'].append(red)
        mkeys = ['precision', 'recall', 'mAP', 'AP', 'acc', 'red']
        res = {mk: [] for mk in mkeys}
        res['ap_class'] = []
        res['loss'] = sum(losses) / len(dataloader)
        for label, class_name in enumerate(self.CLASSES):
            if len(metric_dict_by_class[label]['num_objs']) == 0:
                continue
            num_class_objs = sum(metric_dict_by_class[label]['num_objs'])
            for mk in mkeys:
                res[mk].append(sum(ni * pi for ni, pi in zip(metric_dict_by_class[label]['num_objs'],
                                                             metric_dict_by_class[label][mk])) / num_class_objs)
            res['ap_class'].append(label)
        return res

    def compute_loss(self, model_output, *args, **kwargs):
        return sum(list(model_output.values()))

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)

def iou(real_box, pred_box):
    """Calculates Intersection over Union for real and predicted bounding box."""

    if len(real_box) == 0:
        raise "Length of target box is 0."
    if len(pred_box) == 0:
        raise "Length of predicted box is 0."
    x1 = max(real_box[0], pred_box[0])
    y1 = max(real_box[1], pred_box[1])
    x2 = min(real_box[2], pred_box[2])
    y2 = min(real_box[3], pred_box[3])
    intersection_area = max(0, x2 - x1 + 1) * max(0, y2 - y1 + 1)
    real_box_area = (real_box[2] - real_box[0] + 1) * (real_box[3] - real_box[1] + 1)
    pred_box_area = (pred_box[2] - pred_box[0] + 1) * (pred_box[3] - pred_box[1] + 1)
    return intersection_area / (real_box_area + pred_box_area - intersection_area)


def precision_recall_ap(targets, predictions, iou_thresh=0.5, class_label=None):
    """Calculates Precision, Recall and Average Precision for given threshold."""

    tp_list = []
    num_gt_boxes = 0
    for target in targets:
        for idx, box in enumerate(target['boxes']):
            if target['labels'][idx] == class_label or class_label is None:
                num_gt_boxes += 1
    for target, pred in zip(targets, predictions):
        if len(pred['boxes']) == 0:
            continue
        if len(target['boxes']) == 0:
            for i in range(len(pred['boxes'])):
                tp_list.append(0)
        for target_idx, target_box in enumerate(target['boxes']):
            if target['labels'][target_idx] != class_label and class_label is not None:
                continue
            is_append = False
            for pred_box in pred['boxes']:
                if iou(target_box, pred_box) > iou_thresh:
                    tp_list.append(1)
                    is_append = True
                    break
            tp_list.append(0) if not is_append else None
    if len(tp_list) == 0:
        if num_gt_boxes == 0:
            return 1, 1, 1, num_gt_boxes
        return 0, 0, 0, num_gt_boxes
    if num_gt_boxes == 0:
        return 0, 0, 0, num_gt_boxes
    tp_list.sort(reverse=True)
    precision_values = [sum(tp_list[:i+1])/len(tp_list[:i+1]) for i in range(len(tp_list))]
    recall_values = [sum(tp_list[:i+1])/num_gt_boxes for i in range(len(tp_list))]
    precision, recall = precision_values[-1], recall_values[-1]
    average_precision = precision if len(precision_values) == 1 else auc(recall_values, precision_values)
    return precision, recall, average_precision, num_gt_boxes

def f1_score(precision, recall):
    """Calculates F1-score."""

    return 0 if precision + recall == 0 else 2 * (precision * recall) / (precision + recall)

def mean_average_precision(targets, predictions, class_label=None, thresh_values=np.arange(0.5, 1, 0.05)):
    """Calculates mean Average Precision."""

    ap_list = []
    for thresh in thresh_values:
        ap_list.append(precision_recall_ap(targets, predictions, thresh, class_label)[2])
    return np.mean(ap_list)


def accuracy(targets, predictions, iou_thresh=0.5, class_label=None):
    """Calculates batch accuracy."""

    num_correct_predictions = 0
    num_predictions = 0
    for pred in predictions:
        for idx, box in enumerate(pred['boxes']):
            if pred['labels'][idx] == class_label or class_label is None:
                num_predictions += 1
    for target, pred in zip(targets, predictions):
        if len(target['boxes']) == 0 or len(pred['boxes']) == 0:
            continue
        for target_idx, target_box in enumerate(target['boxes']):
            if target['labels'][target_idx] != class_label and class_label is not None:
                continue
            for pred_idx, pred_box in enumerate(pred['boxes']):
                if iou(target_box, pred_box) > iou_thresh \
                        and target['labels'][target_idx] == pred['labels'][pred_idx]:
                    num_correct_predictions += 1
                    break
    return 1 if num_predictions == 0 else num_correct_predictions / num_predictions


def redundancy(targets, predictions, class_label=None):
    """Calculates redundancy of predictions."""

    num_predictions, num_gt_boxes = 0, 0
    for pred in predictions:
        for idx, box in enumerate(pred['boxes']):
            if pred['labels'][idx] == class_label or class_label is None:
                num_predictions += 1
    for target in targets:
        for idx, box in enumerate(target['boxes']):
            if target['labels'][idx] == class_label or class_label is None:
                num_gt_boxes += 1
    return num_predictions if num_gt_boxes == 0 else num_predictions / num_gt_boxes



from torch import nn
import torch.nn.functional as F
from flgo.utils.fmodule import FModule


class Model(FModule):
    def __init__(self):
        super(Model, self).__init__()
        self.encoder = nn.Sequential(
            nn.Flatten(1),
            nn.Linear(784, 200),
            nn.ReLU(),
            nn.Linear(200, 200),
            nn.ReLU()
        )
        self.head = nn.Linear(200, 10)


    def forward(self, x):
        x = self.encoder(x)
        x = self.head(x)
        return x



def init_local_module(object):
    pass

def init_global_module(object):
    object.model = Model().to(object.device) if 'Server' in object.__class__.__name__ else None


from torch import nn
from flgo.utils.fmodule import FModule

class Model(FModule):
    def __init__(self, dim_in = 3, dim_out = 2):
        super(Model, self).__init__()
        self.layer = nn.Linear(dim_in, dim_out)
        self.layer.bias.data.zero_()
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        x = self.layer(x)
        return x

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


import torchvision
import os
import torch.nn as nn
import torch.nn.functional as F
import flgo.benchmark

transform = torchvision.transforms.Compose(
    [torchvision.transforms.ToTensor(),
     torchvision.transforms.Normalize((0.4914, 0.4822, 0.4465), (0.247, 0.243, 0.262))]
)

path = os.path.join(flgo.benchmark.data_root,  'CIFAR100')
train_data = torchvision.datasets.CIFAR100(root=path, train=True, download=True, transform=transform)
test_data = torchvision.datasets.CIFAR100(root=path, train=False, download=True, transform=transform)

class Model(nn.Module):
    def __init__(self):
        super(Model, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=3, out_channels=64, kernel_size=5)
        self.conv2 = nn.Conv2d(in_channels=64, out_channels=64, kernel_size=5)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.fc1 = nn.Linear(1600, 384)
        self.fc2 = nn.Linear(384, 192)
        self.fc3 = nn.Linear(192, 100)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 1600)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

def get_model():
    return Model()



from flgo.experiment.logger import BasicLogger
import numpy as np
import flgo.simulator.base as ss

class DecLogger(BasicLogger):
    r"""Simple Logger. Only evaluating model performance on testing dataset and validation dataset."""
    def initialize(self):
        """This method is used to record the stastic variables that won't change across rounds (e.g. local_movielens_recommendation data size)"""
        for c in self.participants:
            self.output['client_datavol'].append(len(c.train_data))

    """This logger only records metrics on validation dataset"""
    def log_once(self, *args, **kwargs):
        local_val_metrics = []
        local_test_metrics = []
        for c in self.participants:
            local_val_metrics.append(c.test(flag = 'val'))
            local_test_metrics.append(c.test(flag='test'))
        val_met_name = list(local_val_metrics[0].keys()) if local_val_metrics[0] is not None else []
        if len(val_met_name)>0:
            val_met = {met_name: [] for met_name in val_met_name}
            for cid in range(len(self.participants)):
                for k in val_met_name:
                    val_met[k].append(local_val_metrics[cid][k])
            for k in val_met_name:
                self.output['val_'+k].append(np.mean(val_met[k]))
        test_met_name = list(local_test_metrics[0].keys()) if local_test_metrics[0] is not None else []
        if len(test_met_name)>0:
            test_met = {met_name: [] for met_name in test_met_name}
            for cid in range(len(self.participants)):
                for k in test_met_name:
                    test_met[k].append(local_test_metrics[cid][k])
            for k in test_met_name:
                self.output['test_'+k].append(np.mean(test_met[k]))
        # output to stdout
        self.show_current_output()